/*
 * emptris - Tetris clone for empeg car.
 * 
 * 	Copyright (C) 2000 Brian Mihulka <bmihulka@hulkster.net>
 *	Portions Copyright (C) 2002 Trevor Man <tman_empeg@trejan.com>
 *	Small fork() update by John Heathco <john@heathcosoft.com>
 * 
 * 	This program is free software; you can redistribute it and/or modify
 * 	it under the terms of the GNU General Public License as published by
 * 	the Free Software Foundation; either version 2 of the License, or
 * 	(at your option) any later version.
 * 
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 * 
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software Foundation,
 * 	Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * 		History:
 * 			Thr Oct 19 2000: Initial public release
 *			Tue Jul 24 2001: Remote fix for empeg version 1.03
 *			Sat Aug 04 2001: Added support for rio remote
 *			Fri Aug 02 2002: Modified to bind into Hijack menu - TSM
 *			Fri Aug 02 2002: Add fork() function
 */

#define EMPTRIS_VERSION "emptris v1.1"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <sys/mman.h>

#include <sys/ioctl.h>

#include <stdlib.h>
#include <sys/time.h>

#include <stdio.h>
#include <errno.h>

#define EMPEG_DISPLAY_MAGIC             'd'

#define EMPEG_HIJACK_WAITMENU           _IO(EMPEG_DISPLAY_MAGIC, 80)
#define EMPEG_HIJACK_DISPWRITE          _IO(EMPEG_DISPLAY_MAGIC, 82)
#define EMPEG_HIJACK_BINDBUTTONS        _IO(EMPEG_DISPLAY_MAGIC, 83)
#define EMPEG_HIJACK_DISPCLEAR          _IO(EMPEG_DISPLAY_MAGIC, 86)
#define EMPEG_HIJACK_DISPTEXT           _IO(EMPEG_DISPLAY_MAGIC, 87)
#define EMPEG_HIJACK_POLLBUTTONS        _IO(EMPEG_DISPLAY_MAGIC, 89)
#define EMPEG_SCREEN_ROWS               32
#define EMPEG_SCREEN_COLS               128
#define EMPEG_SCREEN_BYTES              (EMPEG_SCREEN_ROWS * EMPEG_SCREEN_COLS / 2)

#define IR_TOP_BUTTON_PRESSED           0x00000000
#define IR_RIGHT_BUTTON_PRESSED         0x00000002
#define IR_LEFT_BUTTON_PRESSED          0x00000004
#define IR_BOTTOM_BUTTON_PRESSED        0x00000006
#define IR_KNOB_PRESSED                 0x00000008
#define IR_KNOB_RIGHT                   0x0000000a
#define IR_KNOB_LEFT                    0x0000000b

#define IR_KW_1_PRESSED                 0x00b94601
#define IR_KW_PREVTRACK_PRESSED         0x00b9460A
#define IR_KW_NEXTTRACK_PRESSED         0x00b9460B
#define IR_KW_VOL_PLUS_PRESSED          0x00b94614
#define IR_KW_DNPP_PRESSED              0x00b9465E

#define IR_RIO_1_PRESSED                0x0020df00
#define IR_RIO_PREVTRACK_PRESSED        0x0020df10
#define IR_RIO_NEXTTRACK_PRESSED        0x0020df11
#define IR_RIO_MENU_PRESSED             0x0020df12
#define IR_RIO_VOLPLUS_PRESSED          0x0020df13

void draw_tile(int **d, int x, int y, int c);
void erase_tile(int **d, int x, int y);
void refresh_board(int **d, int **g);
void new_piece(int **piece, int num);
void rotate_piece(int **piece, int num, int *rot);
void update_display(int fd, int **d, unsigned char *screenbuf);
void set_pixel(int **d, int x, int y, int level);
void draw_box(int **d, int x1, int y1, int x2, int y2, int level);

int main(void)
{
  int fd;
  int **display;
  int **gameboard;
  int **piece, **newpiece;
  int i, j, k;
  int test, rnd, end, line, rot, newrot, mvtst, drop;
  int delay = 13000, lines, prevline;
  char buf[128];
  unsigned long data;
  unsigned buttons[16] = { 16,
    // Drop button
    IR_BOTTOM_BUTTON_PRESSED,
    IR_KW_DNPP_PRESSED,
    IR_RIO_MENU_PRESSED,

    // Move left button
    IR_LEFT_BUTTON_PRESSED,
    IR_KW_PREVTRACK_PRESSED,
    IR_RIO_PREVTRACK_PRESSED,

    // Move right button
    IR_RIGHT_BUTTON_PRESSED,
    IR_KW_NEXTTRACK_PRESSED,
    IR_RIO_NEXTTRACK_PRESSED,

    // Rotate button
    IR_KW_VOL_PLUS_PRESSED,
    IR_RIO_VOLPLUS_PRESSED,
    IR_TOP_BUTTON_PRESSED,

    // Quit button
    IR_KNOB_PRESSED,
    IR_KW_1_PRESSED,
    IR_RIO_1_PRESSED
  };
  unsigned char screenbuf[EMPEG_SCREEN_BYTES] = { 0 };
  const char *hijack_menu[] = { "emptris", NULL };

  srand(time(0));

  display = calloc(128, sizeof(int *));
  for (i = 0; i < 128; i++) {
    display[i] = calloc(32, sizeof(int));
  }
  gameboard = calloc(44, sizeof(int *));
  for (i = 0; i < 44; i++) {
    gameboard[i] = calloc(10, sizeof(int));
  }
  piece = calloc(5, sizeof(int *));
  for (i = 1; i < 5; i++) {
    piece[i] = calloc(3, sizeof(int));
  }
  newpiece = calloc(5, sizeof(int *));
  for (i = 1; i < 5; i++) {
    newpiece[i] = calloc(3, sizeof(int));
  }

  fprintf(stderr, EMPTRIS_VERSION " starting...\n");

  if ((fd = open("/dev/display", O_RDWR)) == -1) {
    fprintf(stderr, "Unable to open display.\n");
    exit(1);
  }

  if( fork( ) != 0 )
    return 0;
  
  while (1) {
    // Reset line count
    lines = 0;
    prevline = 0;

    // Reset drop flag
    drop = 0;

    // Reset quit flag
    end = 0;
  
    // Clear board
    for (i = 0; i < 44; i++) {
      memset(gameboard[i], 0, 10 * sizeof(int));
    }

    // Clear display
    for (i = 0; i < 128; i++) {
      memset(display[i], 0, 32 * sizeof(int));
      display[i][0] = 3;
      display[i][31] = 3;
    }
    for (i = 0; i < 32; i++) {
      display[127][i] = 3;
    }

    if (ioctl(fd, EMPEG_HIJACK_WAITMENU, hijack_menu) == -1) {
      fprintf(stderr, "Unable to bind into Hijack menu.\n");
      return 1;
    }
    if (ioctl(fd, EMPEG_HIJACK_BINDBUTTONS, buttons) == -1) {
      fprintf(stderr, "Unable to bind buttons.\n");
      return 1;
    }

    if (ioctl(fd, EMPEG_HIJACK_DISPCLEAR) == -1) {
      fprintf(stderr, "Unable to clear screen.\n");
      return 1;
    }

    refresh_board(display, gameboard);
    update_display(fd, display, screenbuf);

    rnd = (int) (7.0 * rand() / (RAND_MAX));
    new_piece(piece, rnd);

    do {
      rot = test = 0;
      for (i = 1; i < 5; i++) {
	draw_tile(display, piece[i][1], piece[i][2], rnd % 2 + 1);
      }
      update_display(fd, display, screenbuf);
      do {
	if (!drop)
	  usleep(delay);

	if (ioctl(fd, EMPEG_HIJACK_POLLBUTTONS, &data) != -1) {
	  switch (data) {
	  case IR_BOTTOM_BUTTON_PRESSED:
	  case IR_KW_DNPP_PRESSED:
	  case IR_RIO_MENU_PRESSED:
	    drop = 1;
	    break;

	  case IR_LEFT_BUTTON_PRESSED:
	  case IR_KW_PREVTRACK_PRESSED:
	  case IR_RIO_PREVTRACK_PRESSED:
	    if (!
		(piece[1][2] == 10 || piece[2][2] == 10
		 || piece[3][2] == 10 || piece[4][2] == 10)) {
	      mvtst = 0;
	      for (i = 1; i < 5; i++) {
		if (gameboard[piece[i][1]][piece[i][2] + 1])
		  mvtst++;
	      }
	      if (!mvtst) {
		for (i = 1; i < 5; i++) {
		  draw_tile(display, piece[i][1], piece[i][2], 0);
		  piece[i][2]++;
		}
	      }
	    }
	    break;

	  case IR_RIGHT_BUTTON_PRESSED:
	  case IR_KW_NEXTTRACK_PRESSED:
	  case IR_RIO_NEXTTRACK_PRESSED:
	    if (!
		(piece[1][2] == 1 || piece[2][2] == 1 || piece[3][2] == 1
		 || piece[4][2] == 1)) {
	      mvtst = 0;
	      for (i = 1; i < 5; i++) {
		if (gameboard[piece[i][1]][piece[i][2] - 1])
		  mvtst++;
	      }
	      if (!mvtst) {
		for (i = 1; i < 5; i++) {
		  draw_tile(display, piece[i][1], piece[i][2], 0);
		  piece[i][2]--;
		}
	      }
	    }
	    break;

	  case IR_KW_VOL_PLUS_PRESSED:
	  case IR_RIO_VOLPLUS_PRESSED:
	  case IR_TOP_BUTTON_PRESSED:
	    for (i = 1; i < 5; i++) {
	      newpiece[i][1] = piece[i][1];
	      newpiece[i][2] = piece[i][2];
	    }
	    newrot = rot;
	    rotate_piece(newpiece, rnd, &newrot);
	    while (newpiece[1][2] < 1 || newpiece[2][2] < 1
		   || newpiece[3][2] < 1 || newpiece[4][2] < 1) {
	      for (i = 1; i < 5; i++) {
		newpiece[i][2]++;
	      }
	    }
	    while (newpiece[1][2] > 10 || newpiece[2][2] > 10
		   || newpiece[3][2] > 10 || newpiece[4][2] > 10) {
	      for (i = 1; i < 5; i++) {
		newpiece[i][2]--;
	      }
	    }
	    mvtst = 0;
	    for (i = 1; i < 5; i++) {
	      if (newpiece[i][1] < 1)
		mvtst++;
	      if (!mvtst) {
		if (gameboard[newpiece[i][1]][newpiece[i][2]])
		  mvtst++;
	      }
	    }
	    if (!mvtst) {
	      for (i = 1; i < 5; i++) {
		draw_tile(display, piece[i][1], piece[i][2], 0);
		piece[i][1] = newpiece[i][1];
		piece[i][2] = newpiece[i][2];
	      }
	      rot = newrot;
	    }
	    break;

	  case IR_KNOB_PRESSED:
	  case IR_KW_1_PRESSED:
	  case IR_RIO_1_PRESSED:
	    goto quit;
	  }
	}
	for (i = 1; i < 5; i++) {
	  draw_tile(display, piece[i][1], piece[i][2], rnd % 2 + 1);
	}
	update_display(fd, display, screenbuf);
	if (!drop)
	  usleep(delay);

	for (i = 4; i > 0; i--) {
	  draw_tile(display, piece[i][1], piece[i][2], 0);
	}

	for (i = 1; i < 5; i++) {
	  if (piece[i][1] == 42)
	    test++;
	  if (gameboard[piece[i][1] + 1][piece[i][2]])
	    test++;
	}

	if (!test) {
	  for (i = 1; i < 5; i++) {
	    piece[i][1]++;
	  }
	}

	for (i = 1; i < 5; i++) {
	  draw_tile(display, piece[i][1], piece[i][2], rnd % 2 + 1);
	}
	update_display(fd, display, screenbuf);

	if (!drop)
	  usleep(delay);

      } while (!test);

      drop = 0;

      for (i = 1; i < 5; i++) {
	gameboard[piece[i][1]][piece[i][2]] = rnd % 2 + 1;
      }

      for (i = 1; i < 43; i++) {
	line = 0;
	for (j = 1; j < 11; j++) {
	  if (gameboard[i][j])
	    line++;
	}
	if (line == 10) {
	  lines++;
	  for (j = i; j > 1; j--) {
	    for (k = 1; k < 11; k++) {
	      gameboard[j][k] = gameboard[j - 1][k];
	    }
	  }
	  for (k = 1; k < 11; k++) {
	    gameboard[1][k] = 0;
	  }
	  refresh_board(display, gameboard);
	  update_display(fd, display, screenbuf);
	}
      }

      if ((lines - prevline) / 10) {
	prevline = lines;
	if (delay > 5000) {
	  delay -= 1000;
	}
      }

      rnd = (int) (7.0 * rand() / (RAND_MAX));
      new_piece(piece, rnd);

      for (i = 1; i < 5; i++) {
	if (gameboard[piece[i][1] + 1][piece[i][2]])
	  end++;
      }

    } while (!end);
quit:

    if (ioctl(fd, EMPEG_HIJACK_DISPCLEAR) == -1) {
      fprintf(stderr, "Unable to clear display.\n");
      return 1;
    }
    sprintf(buf, "\n       You cleared %d lines", lines);
    if (ioctl(fd, EMPEG_HIJACK_DISPTEXT, buf) == -1) {
      fprintf(stderr, "Unable to display text.\n");
      return 1;
    }
    sleep(1);
  }

  fprintf(stderr, "Reached end of main.  Shouldn't happen!\n");

  return 1;
}

void draw_tile(int **d, int x, int y, int c)
{
  switch (c) {
  case 0:
    draw_box(d, (x * 3) - 2, (y * 3) - 2, x * 3, y * 3, 0);
    set_pixel(d, (x * 3) - 1, (y * 3) - 1, 0);
    break;
  case 1:
    draw_box(d, (x * 3) - 2, (y * 3) - 2, x * 3, y * 3, 3);
    set_pixel(d, (x * 3) - 1, (y * 3) - 1, 2);
    break;
  case 2:
    draw_box(d, (x * 3) - 2, (y * 3) - 2, x * 3, y * 3, 2);
    set_pixel(d, (x * 3) - 1, (y * 3) - 1, 3);
    break;
  }
}

void erase_tile(int **d, int x, int y)
{
  draw_box(d, (x * 3) - 2, (y * 3) - 2, x * 3, y * 3, 0);
  set_pixel(d, (x * 3) - 1, (y * 3) - 1, 0);
}

void refresh_board(int **d, int **g)
{
  int i, j;

  for (i = 1; i < 43; i++) {
    for (j = 1; j < 11; j++) {
      draw_tile(d, i, j, g[i][j]);
    }
  }
}

void new_piece(int **piece, int num)
{
  switch (num) {
  case 0:
    piece[1][1] = 1;
    piece[1][2] = 5;
    piece[2][1] = 1;
    piece[2][2] = 6;
    piece[3][1] = 2;
    piece[3][2] = 5;
    piece[4][1] = 2;
    piece[4][2] = 6;
    break;
  case 1:
    piece[1][1] = 1;
    piece[1][2] = 5;
    piece[2][1] = 1;
    piece[2][2] = 6;
    piece[3][1] = 1;
    piece[3][2] = 7;
    piece[4][1] = 2;
    piece[4][2] = 6;
    break;
  case 2:
    piece[1][1] = 1;
    piece[1][2] = 4;
    piece[2][1] = 1;
    piece[2][2] = 5;
    piece[3][1] = 1;
    piece[3][2] = 6;
    piece[4][1] = 1;
    piece[4][2] = 7;
    break;
  case 3:
    piece[1][1] = 1;
    piece[1][2] = 5;
    piece[2][1] = 1;
    piece[2][2] = 6;
    piece[3][1] = 1;
    piece[3][2] = 7;
    piece[4][1] = 2;
    piece[4][2] = 5;
    break;
  case 4:
    piece[1][1] = 1;
    piece[1][2] = 4;
    piece[2][1] = 1;
    piece[2][2] = 5;
    piece[3][1] = 1;
    piece[3][2] = 6;
    piece[4][1] = 2;
    piece[4][2] = 6;
    break;
  case 5:
    piece[1][1] = 1;
    piece[1][2] = 5;
    piece[2][1] = 1;
    piece[2][2] = 6;
    piece[3][1] = 2;
    piece[3][2] = 4;
    piece[4][1] = 2;
    piece[4][2] = 5;
    break;
  case 6:
    piece[1][1] = 1;
    piece[1][2] = 4;
    piece[2][1] = 1;
    piece[2][2] = 5;
    piece[3][1] = 2;
    piece[3][2] = 5;
    piece[4][1] = 2;
    piece[4][2] = 6;
    break;
  }
}

void rotate_piece(int **piece, int num, int *rot)
{
  switch (num) {
  case 0:
    break;
  case 1:
    switch (*rot) {
    case 0:
      piece[1][1]--;
      piece[1][2]++;
      piece[3][1]++;
      piece[3][2]--;
      piece[4][1]--;
      piece[4][2]--;
      *rot = 1;
      break;
    case 1:
      piece[1][1]++;
      piece[1][2]++;
      piece[3][1]--;
      piece[3][2]--;
      piece[4][1]--;
      piece[4][2]++;
      *rot = 2;
      break;
    case 2:
      piece[1][1]++;
      piece[1][2]--;
      piece[3][1]--;
      piece[3][2]++;
      piece[4][1]++;
      piece[4][2]++;
      *rot = 3;
      break;
    case 3:
      piece[1][1]--;
      piece[1][2]--;
      piece[3][1]++;
      piece[3][2]++;
      piece[4][1]++;
      piece[4][2]--;
      *rot = 0;
      break;
    }
    break;
  case 2:
    switch (*rot) {
    case 0:
      piece[1][1] -= 2;
      piece[1][2] += 2;
      piece[2][1]--;
      piece[2][2]++;
      piece[4][1]++;
      piece[4][2]--;
      *rot = 1;
      break;
    case 1:
      piece[1][1] += 2;
      piece[1][2] -= 2;
      piece[2][1]++;
      piece[2][2]--;
      piece[4][1]--;
      piece[4][2]++;
      *rot = 0;
      break;
    }
    break;
  case 3:
    switch (*rot) {
    case 0:
      piece[1][1]--;
      piece[1][2]++;
      piece[3][1]++;
      piece[3][2]--;
      piece[4][1] -= 2;
      *rot = 1;
      break;
    case 1:
      piece[1][1]++;
      piece[1][2]++;
      piece[3][1]--;
      piece[3][2]--;
      piece[4][2] += 2;
      *rot = 2;
      break;
    case 2:
      piece[1][1]++;
      piece[1][2]--;
      piece[3][1]--;
      piece[3][2]++;
      piece[4][1] += 2;
      *rot = 3;
      break;
    case 3:
      piece[1][1]--;
      piece[1][2]--;
      piece[3][1]++;
      piece[3][2]++;
      piece[4][2] -= 2;
      *rot = 0;
      break;
    }
    break;
  case 4:
    switch (*rot) {
    case 0:
      piece[1][1]--;
      piece[1][2]++;
      piece[3][1]++;
      piece[3][2]--;
      piece[4][2] -= 2;
      *rot = 1;
      break;
    case 1:
      piece[1][1]++;
      piece[1][2]++;
      piece[3][1]--;
      piece[3][2]--;
      piece[4][1] -= 2;
      *rot = 2;
      break;
    case 2:
      piece[1][1]++;
      piece[1][2]--;
      piece[3][1]--;
      piece[3][2]++;
      piece[4][2] += 2;
      *rot = 3;
      break;
    case 3:
      piece[1][1]--;
      piece[1][2]--;
      piece[3][1]++;
      piece[3][2]++;
      piece[4][1] += 2;
      *rot = 0;
      break;
    }
    break;
  case 5:
    switch (*rot) {
    case 0:
      piece[2][1]++;
      piece[2][2]--;
      piece[3][1] -= 2;
      piece[4][1]--;
      piece[4][2]--;
      *rot = 1;
      break;
    case 1:
      piece[2][1]--;
      piece[2][2]++;
      piece[3][1] += 2;
      piece[4][1]++;
      piece[4][2]++;
      *rot = 0;
      break;
    }
    break;
  case 6:
    switch (*rot) {
    case 0:
      piece[1][1]--;
      piece[1][2]++;
      piece[3][1]--;
      piece[3][2]--;
      piece[4][2] -= 2;
      *rot = 1;
      break;
    case 1:
      piece[1][1]++;
      piece[1][2]--;
      piece[3][1]++;
      piece[3][2]++;
      piece[4][2] += 2;
      *rot = 0;
      break;
    }
    break;
  }
}

void set_pixel(int **d, int x, int y, int level)
{
  d[x][y] = level;
}

void draw_box(int **d, int x1, int y1, int x2, int y2, int level)
{
  int i;

  if (x2 < x1) {
    i = x2;
    x2 = x1;
    x1 = i;
  }
  for (i = x1; i <= x2; i++) {
    set_pixel(d, i, y1, level);
    set_pixel(d, i, y2, level);
  }
  if (y2 < y1) {
    i = y2;
    y2 = y1;
    y1 = i;
  }
  for (i = y1; i <= y2; i++) {
    set_pixel(d, x1, i, level);
    set_pixel(d, x2, i, level);
  }
}

void update_display(int fd, int **d, unsigned char *screenbuf)
{
  int i, j;

  for (i = 63; i >= 0; i--) {
    for (j = 31; j >= 0; j--) {
      screenbuf[i + j * 64] = d[2 * i][j] + (d[2 * i + 1][j] << 4);
    }
  }
  if (ioctl(fd, EMPEG_HIJACK_DISPWRITE, screenbuf) == -1) {
    fprintf(stderr, "Unable to update display.\n");
    exit(1);
  }
}
